// Client.java

/*
 * Rishi Kacker and Matt Pauker
 * Started: 05/12/2001
 */ 

import java.util.*;
import org.w3c.dom.*;

public class Client {
	
	private ArrayList registrations;
	
	private String name;
	
	private boolean dump;
	
	/**
	 * Larger numbers for faster completion of tasks
	 */
	private int speed;
	
	/**
	 * 1/withdrawl represents the probabilty of witdrawing in that period
	 * (1 takes out money every time period)
	 */
	private int withdrawl;
	
	/**
	 * Value between 0 and 1 representing the likelihood of cheating
	 */
	private double cheater;
	
	private Supervisor supervisor;
	private int wealth;
	private int speed_left;
	
	public Client(Supervisor supervisor, int speed, double cheater, int withdrawl, boolean dump, String name) {
		registrations = new ArrayList();
		this.speed = speed;
		
		if (withdrawl <= 1)
			this.withdrawl = 1;
		else
			this.withdrawl = withdrawl;
		
		this.cheater = cheater;
		this.supervisor = supervisor;
		wealth = 0;
		this.dump = dump;
		this.name = name;
	}

	public void AddRegistration(Registration r) {
		registrations.add(r);
	}
	
	public ArrayList GetRegistrations() {
		return registrations;
	}
	
	//return sum of tasks left
	public int Execute() {
		speed_left = speed;
		int sum = 0;
		boolean cashout;
		
		//probalistically decide to witdraw from 0 or all accounts
		cashout = Simulator.GetRandom().nextDouble() < (double)1/withdrawl;
		for (int i = 0; i < registrations.size(); i++) {
			sum += ((Registration)registrations.get(i)).ExecuteTask();
			if (cashout)
				wealth += ((Registration)registrations.get(i)).Withdraw();
		}
		
		return sum;
	}
	
	public boolean IsDumpable() {return dump;}
	
	public int GetSpeed() {
		
		switch(Settings.GetActivityMode()) {
			case Settings.VARIABLE_RESOURCES:
			case Settings.VARIABLE_RESOURCES_MAX:
				return speed;
				/*
				if (speed_left == 0) return 0;
				int s;
				if (Simulator.GetRandom().nextInt(2) == 0) {
					s = (int)Math.ceil(speed/registrations.size());
				} else {
					s = (int)Math.floor(speed/registrations.size());
				}
				if (s >= speed_left) {
					speed_left = 0;
					return speed_left;
				} else {
					speed_left -= s;
					return s;
				} */
			case Settings.CONSTANT_RESOURCES:
			default:
				return Settings.CONSTANT_SPEED;
		}
	}

	public double AverageTasksLeft() {return supervisor.AverageTasksLeft();}
	
	// XML Tag names
	private final static String CLIENT = "client";
	private final static String SPEED = "speed";
	private final static String WITHDRAWL = "withdrawl";
	private final static String CHEATER = "cheater";
	private final static String SPEED_LEFT = "speed_left";
	private final static String WEALTH = "wealth";
	private final static String REGISTRATIONS = "registrations";
	private final static String DUMP = "dump";
	private final static String NAME = "name";
	
	public Element Write(Document doc) {
		Element c_node = doc.createElement(CLIENT);
		
		c_node.setAttribute(SPEED, Integer.toString(speed));
		c_node.setAttribute(WITHDRAWL, Integer.toString(withdrawl));
		c_node.setAttribute(CHEATER, Double.toString(cheater));
		c_node.setAttribute(SPEED_LEFT, Integer.toString(speed_left));
		c_node.setAttribute(WEALTH, Integer.toString(wealth));
		c_node.setAttribute(REGISTRATIONS, Integer.toString(registrations.size()));
		c_node.setAttribute(DUMP, (dump) ? "true" : "false");
		c_node.setAttribute(NAME, name);
		
		for (int i = 0; i < registrations.size(); i++) {
			c_node.appendChild(((Registration)registrations.get(i)).Write(doc));	
		}
		
		return c_node;
	}
	
	private final static String TOTAL_ACTIVITY = "total_activity";
	private final static String TOTAL_TASKS = "total_tasks";
	
	public Element WriteStats(Document doc) {
		Element c_node = doc.createElement(CLIENT);
		
		double ta = 0.0;
		int tt = 0;
		
		for (int i = 0; i < registrations.size(); i++) {
			ta += ((Registration)registrations.get(i)).GetActivity();
			tt += ((Registration)registrations.get(i)).GetAssigned();
		}
		
		c_node.setAttribute(TOTAL_ACTIVITY, Double.toString(ta));
		c_node.setAttribute(TOTAL_TASKS, Integer.toString(tt));
		
		return c_node;
	}
}
