// Simulator.java

/*
 * Rishi Kacker and Matt Pauker
 * Started: 05/13/2001
 */ 

import java.util.*;
import java.io.*;
import java.util.ArrayList;
import org.w3c.dom.*;

public class Simulator {

	private ArrayList clients;  //fully active clients
	private ArrayList unregistered;
	private ArrayList incompleteTasks;
	private double avgtasks;
	private Supervisor supervisor;
	private static Random rand;
	private Stats stats;
	private int numTasksLeft;
	private boolean dirty;
	
	Simulator() {
		
		stats = new Stats();
		
		clients = new ArrayList();
		unregistered = new ArrayList();
		incompleteTasks = new ArrayList();
		numTasksLeft = Settings.NUM_TASKS;
		
		supervisor = new Supervisor(this);
		
		rand = new Random(5);
		
		// Default settings
		Settings.SetActivityMode(Settings.CONSTANT_RESOURCES);
	 	Settings.SetQMode(Settings.STATIC_DECREASING);
		Settings.SetPaymentMode(Settings.VARIABLE_PAYMENT);
		Settings.SetAgreementMode(Settings.UNANIMOUS_AGREEMENT);
		Settings.SetElasticty(0.5);
		
		dirty = false;
	}

	private void DumpClients(String filename)
	{
		Client c;
		
		XMLHandler out = new XMLHandler(XMLHandler.NEW, "clients");
		for (int i = 0; i < clients.size(); i++)
		{
			c = (Client)clients.get(i);
			out.Append(c.Write(out.GetDocument()));
		}
		out.SaveXML(filename);
	}
	
	private void ReadClients(String filename) {
		
		XMLHandler handle = new XMLHandler(XMLHandler.LOAD,filename);
		NodeList nl = handle.GetNodeList("client");
		
		for (int i = 0; i < nl.getLength(); i++) {
			Node n = nl.item(i);
			NamedNodeMap map = n.getAttributes();
			Client c = new Client(supervisor,Integer.parseInt(map.getNamedItem("speed").getNodeValue()),Double.parseDouble(map.getNamedItem("cheater").getNodeValue()),Integer.parseInt(map.getNamedItem("withdrawl").getNodeValue()),(map.getNamedItem("dump").getNodeValue()).equals("true"),map.getNamedItem("name").getNodeValue());
			
			NodeList child = n.getChildNodes();
			for (int j = 0; j < child.getLength(); j++) {
				NamedNodeMap rmap = ((Node)child.item(j)).getAttributes();
				Registration r = new Registration(c,Integer.parseInt(rmap.getNamedItem("registered").getNodeValue()),Double.parseDouble(rmap.getNamedItem("money").getNodeValue()), Double.parseDouble(rmap.getNamedItem("activity").getNodeValue()));
				supervisor.addRegistration(r);
				c.AddRegistration(r);
			}
			
			clients.add(c);
		}
		
		dirty = true;
	}
	
	private void RunScript(String fileName)
	{
		try
		{
			File f = new File(fileName);
			FileReader reader = new FileReader(f);
			int c;
			StringBuffer s = new StringBuffer();
			
			while ((c = reader.read()) != -1)
			{
				if (c == '\n')
				{
					System.out.println("> " + s.toString());
					ProcessCommand(s.toString());
					s.delete(0, s.length());
				}
				else
					s.append(c);
			}
					
			reader.close();
		}
		catch (FileNotFoundException e)
		{
			System.err.println(e);
		}
		catch (IOException ioe)
		{
			System.err.println(ioe);
		}
	}
	
	private void RunTimePeriod() {
		
		Hashtable timeData = new Hashtable();
		
		if (dirty) {
			supervisor.setNewActivity();
			dirty = false;
		}
		

		// Add new tasks to the supervisor
		ArrayList distTasks = GetTasksToDist();
		if (distTasks.size() == 0)
		{
			System.out.println("All tasks completed.");
			return;
		}
		
		supervisor.addTasks(distTasks);
		timeData.put("numUniqueTasks", new Integer(distTasks.size()));
		
		int n = supervisor.AssignTasks();
		timeData.put("numAssignments", new Integer(n));
		
		// Execute each clients task actions and find avgtasks
		int sum = 0;
		int i;
		
		for (i = 0; i < clients.size(); i++) {
			sum += ((Client)clients.get(i)).Execute();
			if (((Client)clients.get(i)).IsDumpable())
				stats.AppendClientData((Client)clients.get(i));
		}
		
		avgtasks = (double)sum / supervisor.numRegistrations();
		
		//System.out.println("DIFF: " + supervisor.calcdiff());
		//System.exit(1);
		
		supervisor.ProcessResults();
		supervisor.UpdateActivity();

		// Add back tasks to our list that were not completed (i.e. results were
		// not unanimous
		int numReassigned = 0;
		
		for (i = 0; i < distTasks.size(); i++)
		{
			Task t = (Task)distTasks.get(i);
			if (!t.isDone())
			{
				numReassigned++;
				incompleteTasks.add(t);
				numTasksLeft++;
			}
		}
		
		distTasks.clear();
		timeData.put("numReassigned", new Integer(numReassigned));
		timeData.put("amtPaid", new Double(supervisor.getAmtPaid()));
		timeData.put("numPaid", new Integer(supervisor.getNumPaid()));
		stats.AppendTimePeriod(timeData);
				
	}
	
	private void RunComplete()
	{
		int n = 0;
		
		while (numTasksLeft > 0)
		{
			RunTimePeriod();

//System.out.println("TIME PERIOD " + n);
			n++;
		}
		//System.out.println("Ran " + n + " times");
	}

	public static Random GetRandom()
	{
		return rand;
	}

	/* 
	 * Assumes the current value of avgtasks is correct.
	 */
	public double AverageTasksLeft() {return avgtasks;}

	private ArrayList GetTasksToDist()
	{
		ArrayList tasks = new ArrayList();
		int n = 0;
		int i;

		while ((n < (Settings.NUM_TASKS / Settings.NUM_TIME_PERIODS)) && (numTasksLeft > 0))
		{
			i = rand.nextInt(numTasksLeft);
			if (i < incompleteTasks.size()) {
				tasks.add((Task)incompleteTasks.remove(incompleteTasks.size() - 1));
			} else {
				tasks.add(new Task());
			}
			numTasksLeft--;
			n++;
		}
		
		return tasks;
	}
	
	public void ProcessCommand(String line)
	{
		String command;
		ArrayList args = new ArrayList();
		int index;
		
		if (line.equals(""))
			return;
			
		line.trim();
		index = line.indexOf(' ');
		if (index != -1)
		{
			command = line.substring(0, index);
			line = line.substring(index + 1, line.length());
			while ((index = line.indexOf(' ')) != -1)
			{
				args.add(line.substring(0, index));
				line = line.substring(index + 1, line.length());
			}
			args.add(line);
		}
		else
			command = line;
			
		if (command.equals("run"))
		{
			if (args.get(0).equals("complete"))
				RunComplete();
			else
			{
				int n = Integer.valueOf((String)args.get(0)).intValue();
				for (int i = 0; i < n; i++)
					RunTimePeriod();
			}
		}
		else if (command.equals("loadclients"))
		{
			ReadClients((String)args.get(0));
		}
		else if (command.equals("script"))
		{
			RunScript((String)args.get(0));
		}
		else if (command.equals("dump"))
		{
			stats.SaveStats((String)args.get(0));
		}
		else if (command.equals("dumpclients"))
		{
			DumpClients((String)args.get(0));
		}
	}
	
	public void AutoRun(String[] args)
	{
		int start;

		if (Settings.GetActivityMode() != Settings.CONSTANT_RESOURCES)
		{
			Settings.SetElasticty(Double.parseDouble(args[0]));
			start = 1;
		}
		else
			start = 0;
		for (int i = start; i < args.length - 1; i++)
			ReadClients(args[i]);
		RunComplete();
		
		stats.SaveStats(args[args.length - 1]);
	}
	
	public static void main(String[] args)
	{
		Simulator sim = new Simulator();
		
		if (args.length > 0)
		{
			sim.AutoRun(args);
		}
		else
		{
			try
			{
				BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
				String line;
				
				while (true)
				{
					System.out.print("> ");
					line = reader.readLine();
					
					if (line.startsWith("quit"))
						break;
						
					sim.ProcessCommand(line);
				}
			}
					
			catch (IOException e)
			{
				System.err.println(e);
			}
		}
	}
}
