//Supervisor.java

/*
 * Rishi Kacker and Matt Pauker
 * Started: 05/12/2001
 */ 

import java.util.*;

class Supervisor
{
	// Data
	private SortedList regList;
	private ArrayList assignedClientsList;
	private ArrayList taskList;
	private Simulator sim;
	private double c;
	
	// Stats
	private int numTasksReassigned;
	private double compOverhead;
	private double amtPaid;
	private int numPaid;
	
	public Supervisor(Simulator s)
	{
		sim = s;
		regList = new SortedList();
		assignedClientsList = new ArrayList();
		taskList = new ArrayList();
		
		double pnot = 1 - Settings.PARTICIPANT_FRACTION;
		int d = Settings.REGISTRATION_PERIOD;
		double e = Settings.DEFECTING_RATION;
		c = (1/pnot) - Math.sqrt(d / (pnot * (1 + e + d)));
		c = .14886066;
	}
	
	// Adds a list of tasks to be distributed on the next pass
	public void addTasks(ArrayList tasks)
	{
		taskList.addAll(tasks);
	}
	
	// Adds a new registration to our list (i.e. one that just finished the registration period)
	public void addRegistration(Registration r)
	{
		regList.add(r, r.GetActivity());
	}
	
	public int numRegistrations() {
		return regList.size();
	}
	
	public void normalizeActivity() {
		double ta = 0.0;
		for (int i = 0; i < regList.size(); i++) {
			ta += ((Registration)regList.getObject(i)).GetActivity();
		}
		
		if (Math.abs(ta - 1.0) > .001) {
			SortedList newRegList = new SortedList();
			for (int i = 0; i < regList.size(); i++) {
				Registration reg = (Registration)regList.getObject(i);
				reg.normalize(ta);
				newRegList.add(reg,reg.GetActivity());
			}
			//System.out.println("NORMALIZED TO " + ta);
			regList.clear();
			regList = newRegList;
		} else {
			//System.out.println("NOT NORMALIZED TO " + ta);
		}
		
	}
	
	public void UpdateActivity() {
		
		if (Settings.GetActivityMode() != Settings.CONSTANT_RESOURCES)
		{	
			SortedList newRegList = new SortedList();
			for (int i = 0; i < regList.size(); i++)
			{
				Registration r = (Registration)regList.getObject(i);
				r.UpdateActivity();
				newRegList.add(r, r.GetActivity());
			}
			//System.out.println("BEFORE: " + sum_b + " AND AFTER: " + sum_a);
			//System.out.println("UPDATE DIFF: " + diff);
			regList.clear();
			regList = newRegList;
			normalizeActivity();
		}
	}
	
	public void setNewActivity() {
		//System.out.println("YOU CALLED ME");
		SortedList newRegList = new SortedList();
		for (int i = 0; i < regList.size(); i++) {
			Registration r = (Registration)regList.getObject(i);
			if (r.GetActivity() == 0.0)
				r.SetActivity(1.0 / regList.size());
			newRegList.add(r, r.GetActivity());
		}
		regList = newRegList;
		normalizeActivity();
	}
	
	// Returns a list of clients whose registrations were
	// assigned tasks on this pass (implementation-use)
	public ArrayList getAssignedClients()
	{
		return assignedClientsList;
	}
	
	// Assigns all currently available tasks to registrations
	// Returns the number of registrations who got tasks
	public int AssignTasks()
	{
		ArrayList regs;
		Registration r;
		
		for (int i = 0; i < regList.size(); i++)
		{
			r = (Registration)regList.getObject(i);
			r.Restart();
		}

		int numAssignments = 0;
		
		for (int i = 0; i < taskList.size(); i++) {
			regs = chooseRegistrations((Task)taskList.get(i));
			numAssignments += regs.size();
		}
		
		return numAssignments;
	}
	
	// Picks registrations to which to assign the current task		
	private ArrayList chooseRegistrations(Task t)
	{
		ArrayList regs;
		Registration r;
		int mode;
		
		regs = new ArrayList();
		mode = Settings.GetQMode();
		if ((mode == Settings.STATIC_DECREASING) || (mode == Settings.STATIC_DECREASING_DISTINCT))
		{
			int n = numToAssign(mode);
			if (n > regList.size())	// Not enough regs, so...
			{
				for (int i = 0; i < regList.size(); i++)
				{
					regs.add(regList.getObject(i));  // Just assign it to every reg
				}
			}
			else
			{
				for (int i = 0; i < n; i++)
				{
					r = pickRegistration();
					// Make sure we don't assign it twice to the same registration
					// and that this reg can still do tasks
					while (regs.contains(r) || r.NumTasksWanted() == 0)
						r = pickRegistration();
					regs.add(r);
					if (Settings.GetPaymentMode() == Settings.CONSTANT_PAYMENT)
						t.AddRegistration(r, Settings.TASK_PAYMENT);
					else
						t.AddRegistration(r, CalculatePayment(r));
					r.AddTask(t);
				}
			}
		}
		else		// Dynamic distribution
		{
			double p;
			boolean done;
			double q, sum;
			
			done = false;
			sum = 0;
			while (!done)
			{
				r = pickRegistration();
				// Make sure we don't assign it twice to the same registration
				// and that this reg can still do tasks
				while (regs.contains(r) || r.NumTasksWanted() == 0)
					r = pickRegistration();
				regs.add(r);
				if (Settings.GetPaymentMode() == Settings.CONSTANT_PAYMENT)
					t.AddRegistration(r, Settings.TASK_PAYMENT);
				else
					t.AddRegistration(r, CalculatePayment(r));

				r.AddTask(t);
				
				sum += r.GetMoney() + Settings.REGISTRATION_PERIOD * Settings.TASK_PAYMENT;
				q = (regs.size() + Settings.DEFECTING_RATION) /
					(regs.size() + Settings.DEFECTING_RATION + sum);			

				p = sim.GetRandom().nextDouble();
				done = p > q;
			}
		}
		
		return regs;
	}
	
	// Picks a registration to which to assign a task
	private Registration pickRegistration()
	{
		if ((Settings.GetActivityMode() == Settings.VARIABLE_RESOURCES) ||
			(Settings.GetActivityMode() == Settings.VARIABLE_RESOURCES_MAX))
		{
			// Choose based on activity
			double p, sum;
			int i;
			
			p = sim.GetRandom().nextDouble();
			sum = 0;
			for (i = 0; i < regList.size(); i++)
			{
				sum += regList.getValue(i); 
				if (p < sum)
					break;
			}
			if (i == regList.size())
				i--; // ERROR - Shouldn't get here (i hope)
				
			return (Registration)regList.getObject(i);
		}
		else // Settings.CONSTANT_RESOURCES
		{
			int n = sim.GetRandom().nextInt(regList.size());
			return (Registration)regList.getObject(n);
		}
	}
	
	// Determine how many tasks to assign this task to (if we're doing static assignment)
	private int numToAssign(int qType)
	{
		double p, sum;
		int i;
		
		if (qType == Settings.STATIC_DECREASING)
		{
			p = sim.GetRandom().nextDouble();
			sum = 0;
			for (i = 1; i < 20; i++)
			{
				sum += (1 - c) * Math.pow(c, i - 1);
				if (p < sum)
					break;
			}
			return i;
		}
		else // STATIC_DECREASING_DISTINCT
		{
			int s = Settings.MIN_ASSIGNMENTS;
			
			p = sim.GetRandom().nextDouble();
			sum = 0;
			for (i = s; i < 20; i++)
			{
				sum += (1 - c) * Math.pow(c, i - s);
				if (p < sum)
					break;
			}
			return i;
		}
	}
	
	
	// Processes the results 'returned' by all the registrations for each task
	public void ProcessResults()
	{
		Task t;
		ArrayList assigns = new ArrayList();

		amtPaid = 0.0;
		numPaid = 0;
				
		for (int i = 0; i < taskList.size(); i++) {
			
			t = (Task)taskList.get(i);
			if (resultsOK(t.getCurrentAssigns()))
			{

				Assignment a;
				
				assigns.clear();
				assigns.addAll(t.getCurrentAssigns());
				if (t.getPrevAssigns() != null)
					assigns.addAll(t.getPrevAssigns());

				for (int j = 0; j < assigns.size(); j++)
				{
					a = (Assignment)assigns.get(j);
					if (a.answer == true)
					{
						double n = a.reg.Pay(a.payment);
						amtPaid += n;
						if (n > 0.0)
							numPaid++;
					}
					//else
					//	a.reg.Kill();
					// else punish?
				}
				
				t.setDone(true);


			}
			else
			{
				t.restart();	// Do the task again on the next pass
				numTasksReassigned++;
			}

		}


		taskList.clear();
		assignedClientsList.clear();


	}

	// Determines if the results of a task were okay (either by
	// majority or unanimous vote) or whether we need to reassign
	// the task on the next pass
	private boolean resultsOK(ArrayList results)
	{
		int numTrue = 0;
	
		for (int i = 0; i < results.size(); i++) {
			if (((Assignment)results.get(i)).answer)
				numTrue++;
		}

		// Are we requiring a unanimous vote, or is majority enough?
		if (Settings.GetAgreementMode() == Settings.UNANIMOUS_AGREEMENT)
			return (numTrue == results.size());
		else
			return ((numTrue * 2) > results.size());
	}
	
	private double CalculatePayment(Registration r)
	{
		/* Old way
		return Settings.TASK_PAYMENT * 
			Math.sqrt(1.0 - 1.0 / (r.GetMoney() + Settings.REGISTRATION_PERIOD * Settings.TASK_PAYMENT));
		*/

		double deposit = Settings.REGISTRATION_PERIOD * Settings.TASK_PAYMENT;
		double n = deposit / (r.GetMoney() + 1.25 * deposit);
		return Settings.TASK_PAYMENT * (1 - Math.pow(n, 3));
	}
	
	public double getAmtPaid()
	{
		return amtPaid;
	}

	public int getNumPaid() { return numPaid; }
	public double AverageTasksLeft() {return sim.AverageTasksLeft();}	
}
	
